@@ -14,10 +14,23 @@ use boolean;
 sub got_map { +{map @$_, map @$_, @{(pop)}} }
 sub got_seq { [map @$_, @{(pop)}] }
 
+my %escapes = (
+    '"' => '"',
+    '/' => "/",
+    "\\" => "\\",
+    b => "\b",
+    f => "\x12",
+    n => "\n",
+    r => "\r",
+    t => "\t",
+);
+
 sub got_string {
     my $string = pop;
-    # XXX need to decode other string escapes here
-    $string =~ s/\\n/\n/g;
+    $string =~ s/\\(["\/\\bfnrt])/$escapes{$1}/ge;
+    # This handles JSON encoded Unicode surrogate pairs
+    $string =~ s/\\u([0-9a-f]{4})\\u([0-9a-f]{4})/pack "U*", hex("$1$2")/ge;
+    $string =~ s/\\u([0-9a-f]{4})/pack "U*", hex($1)/ge;
     return $string;
 }
 
